#define _XOPEN_SOURCE 500

#include <linux/aquosa/qos_debug.h>

#include "lfs_qtrace.h"
#include "lfs_threads.h"

#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#define QTRACE_SYSCALL	0
#define QTRACE_TASK	1
#define QTRACE_FLUSH	2

#define LEAVE_MASK	0x10000
#define ENTER_MASK	0x20000

#define MAX_QEVENTS	32768

extern struct ptraced_child *ptraced_table[HASH_SIZE];

static int qtracefd = -1;
static int qtrace_wakeup;

#define QTRACE_DEV "/dev/qtrace"

static void qtrace_raw_cmd(unsigned int values[])
{
	int ret;

	if (qtracefd == -1) {
		qos_log_debug("Opening " QTRACE_DEV " for r/w");
		qtracefd = open(QTRACE_DEV, O_RDWR);
		qos_chk_do(qtracefd != -1, err("Could not open " QTRACE_DEV
					       "\n  (try sudo chmod go+rw " QTRACE_DEV ")\n"));
	}
	ret = pwrite(qtracefd, values, 4 * 3, 0);
	if (ret < 0) {
		perror("writing");
		err("writing\n");
	}
}

static void qtrace_flush(void)
{
	unsigned int values[3] = { QTRACE_FLUSH, 0, 0 };
	qtrace_raw_cmd(values);
}

static void qtrace_set_syscalls(void)
{
	int i;

	for (i = 0; i < __NR_syscall_max; i++) {
		if (syscalls[i].track) {
			unsigned int values[3] = { QTRACE_SYSCALL, i, 1 };
			qtrace_raw_cmd(values); 
		}
	}
}

static void qtrace_child(pid_t pid, int enable)
{
	unsigned int values[3] = { QTRACE_TASK, pid, enable };
	struct ptraced_child *pc;

	qos_log_debug("%stracing thread %d", enable ? "" : "un", pid);
	qtrace_raw_cmd(values);
	if (enable) {
		if (find_child(pid))
			err("already traced child\n");

		xcalloc(pc, sizeof(struct ptraced_child));
		pc->pid = pid;

		pc->next = ptraced_table[HASH(pid)];
		xcalloc(pc->log, sizeof(struct log_vec));
		pc->cur = pc->log;
		pc->cur_event.orig_eax = -1;
		ptraced_table[HASH(pid)] = pc;
	}
}

static void qtrace_log_event(unsigned int values[])
{
	struct ptraced_child *pc;

	pc = find_child(values[1]);
	if (!pc) {
		qos_log_debug("Unexpected child %d", values[1]);
		qtrace_child(values[1], 0);
		return;
	}

	if ((values[0] & LEAVE_MASK) || qtrace_wakeup) {
		pc->cur_event.orig_eax = values[0];
		pc->cur_event.tv1.tv_sec = values[2];
		pc->cur_event.tv1.tv_usec = values[3] / 1000;
		log_event(pc, &pc->cur_event);
	}
}

static int timeout(struct timeval *end)
{
	struct timeval tv;

	gettimeofday(&tv, NULL);

	return tv.tv_sec > end->tv_sec || (tv.tv_sec == end->tv_sec &&
		 tv.tv_usec > end->tv_usec);
}

static int qtrace_dump(unsigned long usecs)
{
	unsigned int values[4];
	struct timeval end;
	loff_t rn = (qtrace_wakeup) ? 1 : 0;

	gettimeofday(&end, NULL);
	end.tv_sec += usecs / 1000000;
	end.tv_usec += usecs % 1000000;
	if (end.tv_usec >= 1000000) {
		end.tv_usec -= 1000000;
		end.tv_sec++;
	}

	do {
		while (pread(qtracefd, &values, 16, rn) == 16 && !timeout(&end))
			qtrace_log_event(values);
		usleep(50000);
	} while (!timeout(&end));

	return 0;
}

void qtrace_get_events(unsigned long usecs)
{
	int i;
	typedef void (*sighandler_t)(int);

	if (!usecs)
		err("qtrace wants a timeout!\n");

	gettimeofday(&tracing_start, NULL);

	qtrace_set_syscalls();
	qtrace_flush();

	qos_chk_do(num_threads > 0, exit(-1));

	qos_log_debug("tracing ...");
	for (i = 0; i < num_threads; ++i)
		qtrace_child(threads[i], 1);
	qos_log_debug("... done");

	qtrace_dump(usecs);

	for (i = 0; i < num_threads; ++i)
		qtrace_child(threads[i], 0);
	qos_log_debug("exiting...");
}

void qtrace_call_on_events(transf_cb_t cb, double f, double *re, double *im)
{
	struct event *e;
	double val;

	do_for_each_ptrace_event(e) {
		val = e->tv1.tv_sec * 1000000 + e->tv1.tv_usec;
		cb(val, f, re, im);
	} while_for_each_ptrace_event();
}

void qtrace_setup_wakeup(void)
{
	qtrace_wakeup = 1;
}
