#include <linux/aquosa/qos_debug.h>
#include <aquosa/movavg_lib.h>
#include <linux/aquosa/percpred.h>
#include <linux/aquosa/qos_debug.h>
#include <aquosa/qres_lib.h>

#include "cbs.h"
#include "control.h"
#include "lfs_period.h"
#include "lfs_threads.h"
#include "lfs_power.h"

#include <sys/time.h>
#include <time.h>
#include <stdint.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <pthread.h>
#include <signal.h>
#include <string.h>

#include <math.h>

#define MAX_TASK 50

/// Monitoring / tracing time (us)
static int mon_time_us = 4000 * 1000;

/// Sampling period for LFS++ feedback control loop (us)
static int sampling_period = 1000 * 1000;

#define DEF_XF 110
static int x_factor = DEF_XF;

#define DEF_ULUB 0.9
static double u_lub = DEF_ULUB;

static int updating_speed;

typedef struct __rt_info {
	pid_t *threads[MAX_TASK];
	int num_threads;
	uint32_t t;
	percpred_params_t pred_params;
	percpred_t pred;
	qres_sid_t sid;
	int sample_size;
	int discarded;
	pthread_t lfs_tid;
} rt_info;

static int pid[MAX_TASK];
static rt_info rt_infos;
static int task_index = 0;

static uint32_t t = 0;
static int w = 12;
static int d = 2;

pthread_mutex_t lfs_mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t lfs_started_cond = PTHREAD_COND_INITIALIZER;

static int lfs_started;
static int low_water = 40;
static int high_water = 60;

int variant = 0;
int only_detect = 0;
int include_children = 1;
int enable_power = 0;
int exiting = 0;
const char *tracer = "ptrace";
static const char *events_fname = NULL;
static const char *transf_fname = NULL;

static void cleanup(void)
{
	if (enable_power)
		lfs_shutdown_power();

	if (!only_detect) {
		qos_log_debug("Destroying server %d...",
			      rt_infos.sid);
		qres_destroy_server(rt_infos.sid);
	}
}

/*
 * Returns 0 if going from curr_speed to speed is to be considered safe
 * from a schedulability point of view, taking into account any consideration
 * about safety margins etc.  If it returns 0 the change will be made by 
 * the power daemons.
 * Beware that this can be called concurrently with other functions changing
 * task parameters.
 */
int lfs_validate_speed(unsigned long curr_speed, unsigned long speed)
{
	double scale = (double)curr_speed / speed;
	double new_u;
	float bw;

	if (only_detect)
		bw = 0.0;
	else
		qos_chk_ok_exit(qres_get_bandwidth(rt_infos.sid, &bw));
	new_u = bw * scale;

	qos_log_debug("new_u = %g, u_lub = %g, returning %d",
		      new_u, u_lub, new_u < u_lub);
	return new_u < u_lub;
}

/*
 * Should change budgets as needed, and if speed is decreasing
 * should wait for the budget changes to have taken effect.
 * XXX synchronization might be an issue here: I'm not using a
 * mutex because it wouldn't be enough to prevent the feedback
 * to use a stale read (i.e., a read from an old power setting);
 * if this proves to be a real issue the solution should be to
 * skip the feedback loop immediately following a speed change.
 */
void lfs_update_speed(unsigned long curr_speed, unsigned long speed)
{
	double scale = (double)curr_speed / speed;
	qos_bw_t next_bw;

	if (!only_detect && speed <= curr_speed) {
		qres_params_t params;
		updating_speed = 1;
		qos_chk_ok_exit(qres_get_params(rt_infos.sid, &params));
		next_bw = r2bw(params.Q * scale, rt_infos.t);
		qos_log_debug("new BW %f scale = %lf\n",
			      (float)next_bw/MAX_BW, scale);
		qres_set_bandwidth(rt_infos.sid, next_bw);
		usleep(2*rt_infos.t);
		updating_speed = 0;
	}
}

/** Ideally, detach all threads from servers */
void sigint_handler(int x) {
	qos_log_debug("Got signal, terminating");
	exiting = 1;
}

int real_child_body(rt_info *p) {
	int res;
	qres_time_t prev_spent, curr_spent;
	qres_atime_t curr_time;
	struct timespec next_obs_time;
	qres_time_t exec_time, next_time;

	if (p->t == 0) {
		/* @todo Infer period */
		p->t = infer_period(mon_time_us, variant,
				    events_fname, transf_fname);
		qos_log_debug("Observed %d (us), detected t %d (us), f %g (Hz)",
			      mon_time_us, p->t,
			      p->t > 0 ? round(1000000.0*10/p->t)/10.0 : 0.0);
		qos_chk_do(p->t != 0, pthread_exit((void*) -1));
	}

	pthread_mutex_lock(&lfs_mutex);
	lfs_started = 1;
	pthread_cond_signal(&lfs_started_cond);
	pthread_mutex_unlock(&lfs_mutex);

	if (only_detect)
		pthread_exit(0);

	qos_log_debug("Wrapping threads into reservation with period %u", p->t);
	res = transform_threads(p->t);
	if (res < 0) {
		qos_log_err("Cannot transform tasks in CBS");
		return -1;
	}

	qos_chk_ok_exit(qres_get_sid(pid[0], threads[0], &p->sid));

	p->pred_params = (percpred_params_t) {
		.base = {
			.type = QMGR_PRED_PERC,
		},
		.sample_size = p->sample_size,
		.discarded = p->discarded,
	};
	qos_log_debug("Setting percpred params: sample_size %d  discarded %d",
		      p->pred_params.sample_size, p->pred_params.discarded);
	percpred_init(&p->pred, &p->pred_params);
	qos_chk_ok_do(qres_get_exec_time(p->sid, &prev_spent, &curr_time), pthread_exit(NULL));
	clock_gettime(CLOCK_MONOTONIC, &next_obs_time);

	while (!exiting) {
		next_obs_time.tv_nsec += sampling_period * 1000;
		if (next_obs_time.tv_nsec > 1000000000ul) {
			next_obs_time.tv_nsec -= 1000000000ul;
			next_obs_time.tv_sec++;
		}
		if (clock_nanosleep(CLOCK_MONOTONIC, TIMER_ABSTIME,
				    &next_obs_time, NULL) < 0)
			break;

		/* Compute sensor value */
		qos_chk_ok_do(qres_get_exec_time(p->sid, &curr_spent, &curr_time), pthread_exit(NULL));
		exec_time = curr_spent - prev_spent;
		prev_spent = curr_spent;

		qos_log_debug("Measured exec time = " QRES_TIME_FMT
			      " / %d  task_period = " QRES_TIME_FMT,
			      exec_time, sampling_period, (qres_time_t) p->t);

		/* Normalize measured execution time to actual
		 * application period */
		exec_time = exec_time * p->t / sampling_period;

		/* Compute control action */
		qos_log_debug("Adding computation time " QRES_TIME_FMT,
			      exec_time);
		percpred_add_sample(&p->pred.base, exec_time);
		next_time = percpred_get_exp_value(&p->pred.base);
		/* Provide x_factor% of the measured value */
		qres_time_t next_budget = next_time * x_factor / 100 + 100;
		qos_log_debug(
			"exec_time = " QRES_TIME_FMT
			"  quantile = " QRES_TIME_FMT
			"  next_budget = " QRES_TIME_FMT
			"  period = " QRES_TIME_FMT,
			exec_time, next_time, next_budget, (qres_time_t) p->t);

		if (next_budget > 100 && !updating_speed) {
			/* Apply actuator */
			qos_bw_t next_bw = r2bw(next_budget, p->t);
			qres_set_bandwidth(p->sid, next_bw);
		} else {
			qos_log_debug("Skipping set of too low bandwidth: exec time quantile=" QRES_TIME_FMT, next_time);
		}
	}

	return 0;
}

void *child_body(void *arg) {
	rt_info *p = (rt_info *) arg;
	int exit_code = real_child_body(p);
	return (void *) (unsigned long) exit_code;
}

void usage(void) {
	printf("Options:\n");
	printf("  -h ............ This help message\n");
	printf("  -o ............ Only detect period and exit\n");
	printf("  -c ............ Do not include children\n");
	printf("  -r ............ Enable power management (needs root privileges)\n");
	printf("  -g ............ Enable geometric averages in heuristics 1 and 2\n");
	printf("  -f <fname>..... Write events to <fname>\n");
	printf("  -F <fname>..... Write transform to <fname>\n");
	printf("  -s <period> ... Set period in us of LFS++ control loop\n");
	printf("  -p <pid> ...... Set pid of process to control\n");
	printf("  -t <period> ... Set period of controlled process\n");
	printf("  -T <tracer> ... Set tracer method (ptrace [default], qtrace, qtrace_wakeup)\n");
	printf("  -m <time_us> .. Set monitoring/tracing time (defaults to 4s)\n");
	printf("  -w <samples> .. Set predictor's moving window size (defaults to %d)\n", w);
	printf("  -d <samples> .. Set predictor's discarded param (defaults to %d)\n", d);
	printf("  -v <variant> .. Set period detector variant: 0 [default], 1, 2\n");
	printf("                  0 - Max\n");
	printf("                  1 - Only local peaks higher than threshold*avg, at most k_max multiples\n");
	printf("  -S <threshold>  Set threshold (defaults to %g), use 0.0 to disable\n", peaks_threshold);
	printf("  -K <threshold>  Set k_max (defaults to %d), use 0 to disable\n", k_max);
	printf("  -x <x_factor> . Set x_factor (defaults to %d)\n", DEF_XF);
	printf("  -i <eps_hz> ... Set freq_min (defaults to %g Hz)\n", freq_min);
	printf("  -a <eps_hz> ... Set freq_max (defaults to %g Hz)\n", freq_max);
	printf("  -l <eps_hz> ... Set freq_delta (defaults to %g Hz)\n", freq_delta);
	printf("  -e <eps_hz> ... Set freq_eps (defaults to %g Hz)\n", freq_eps);
	printf("  -L <perc> ..... Set powernowd low water (defaults to %d)\n", low_water);
	printf("  -H <perc> ..... Set powernowd high water (defaults to %d)\n", high_water);
}

int main(int argc, char *argv[])
{
	int v, i, err;

	for(;;) {
		v = getopt(argc, argv, "l:i:a:e:f:F:m:x:p:s:t:T:w:d:rv:S:K:L:H:ohcg");
		if (v == -1) {
			break;
		}

		switch(v) {
		case 'p':
			pid[task_index++] = atoi(optarg);
			break;
		case 't':
			if (task_index == 0) {
				fprintf(stderr, "Invalid Period\n");
				return -1;
			}
			t = atoi(optarg);
			break;
		case 'T':
			tracer = optarg;
			break;
		case 'w':
			w = atoi(optarg);
			break;
		case 'd':
			d = atoi(optarg);
			break;
		case 'v':
			variant = atoi(optarg);
			break;
		case 'S':
			peaks_threshold = strtod(optarg, NULL);
			break;
		case 'K':
			k_max = atoi(optarg);
			break;
		case 'o':
			only_detect = 1;
			break;
		case 'c':
			include_children = 0;
			break;
		case 'r':
			enable_power = 1;
			break;
		case 'g':
			transf_type = TRANSF_GEOMETRIC;
			break;
		case 'f':
			events_fname = strdup(optarg);
			break;
		case 'F':
			transf_fname = strdup(optarg);
			break;
		case 's':
			sampling_period = atoi(optarg);
			break;
		case 'x':
			x_factor = atoi(optarg);
			break;
		case 'm':
			mon_time_us = atoi(optarg);
			break;
		case 'i':
			freq_min = strtod(optarg, NULL);
			break;
		case 'a':
			freq_max = strtod(optarg, NULL);
			break;
		case 'l':
			freq_delta = strtod(optarg, NULL);
			break;
		case 'e':
			freq_eps = strtod(optarg, NULL);
			break;
		case 'L':
			low_water = strtod(optarg, NULL);
			break;
		case 'H':
			high_water = strtod(optarg, NULL);
			break;
		case 'h':
			usage();
			exit(0);
			break;
		}
	}

	if (config_tracer(tracer))
		exit(-1);

	if (!only_detect && cbs_init() < 0)
		exit(-1);

	signal(SIGINT, sigint_handler);

	for (i = 0; i < task_index; ++i) {
		if (include_children)
			qos_chk_do(add_threads(pid[i]) > 0, exit(-1));
		else
			qos_chk_do(add_thread(pid[i]) > 0, exit(-1));
	}
	rt_infos.t = t;
	rt_infos.sample_size = w;
	rt_infos.discarded = d;
	err = pthread_create(&rt_infos.lfs_tid, NULL, child_body, &rt_infos);
	if (err) {
		qos_log_err("Can't start monitor thread\n");
		exit(-1);
	}

	pthread_mutex_lock(&lfs_mutex);
	if (!lfs_started)
		pthread_cond_wait(&lfs_started_cond, &lfs_mutex);
	pthread_mutex_unlock(&lfs_mutex);

	if (enable_power) {
		err = lfs_setup_power(AGGRESSIVE, high_water, low_water);
		if (err)
			qos_log_err("Power management capabilities disabled.");
		enable_power = 0;
		if (only_detect)
			while (!exiting)
				sleep(1);
	}

	pthread_join(rt_infos.lfs_tid, NULL);

	cleanup();

	qos_log_debug("Exiting.\n");
	exit(0);
}
